use crate::{
    token::{Token, TokenKind, TokenValue},
    util::hash,
};
use std::{collections::HashMap, str::Chars};

pub struct Lexer<'a> {
    text: &'a str,
    chars: Chars<'a>,
    current: Option<char>,
    index: usize,
    line: i32,
    col: i32,
    symbols: HashMap<u64, String>,
    keywords: HashMap<String, TokenKind>,
}

impl<'a> Lexer<'a> {
    pub fn new(text: &'a str) -> Self {
        let mut chars = text.chars();
        let current = chars.next();
        Self {
            text,
            chars,
            current,
            index: 0,
            line: 1,
            col: 1,
            symbols: HashMap::new(),
            keywords: Self::make_keywords(),
        }
    }

    fn make_keywords() -> HashMap<String, TokenKind> {
        let mut keywords = HashMap::new();
        keywords.insert("if".to_string(), TokenKind::If);
        keywords.insert("else".to_string(), TokenKind::Else);
        keywords.insert("loop".to_string(), TokenKind::Loop);
        keywords.insert("break".to_string(), TokenKind::Break);
        keywords.insert("let".to_string(), TokenKind::Let);
        keywords.insert("fn".to_string(), TokenKind::Fn);
        keywords.insert("return".to_string(), TokenKind::Return);
        keywords
    }

    pub fn symbols(self) -> HashMap<u64, String> {
        self.symbols
    }

    fn next_token(&mut self) -> Option<Token> {
        match self.current {
            None => None,
            Some(' ' | '\t' | '\n') => {
                self.step();
                self.next_token()
            }
            Some(ch @ ('a'..='z' | 'A'..='Z' | '_')) => {
                let mut value = String::new();
                value.push(ch);
                self.step();
                loop {
                    match self.current {
                        Some(ch @ ('a'..='z' | 'A'..='Z' | '0'..='9' | '_')) => {
                            value.push(ch);
                            self.step();
                        }
                        _ => {
                            if let Some(kind) = self.keywords.get(&value) {
                                return self.token(kind.clone());
                            }
                            let id = hash(&value);
                            self.symbols.insert(id, value);
                            break self.token_with_value(TokenKind::Id, TokenValue::Id(id));
                        }
                    }
                }
            }
            Some(ch @ ('1'..='9')) => {
                let mut value = String::new();
                value.push(ch);
                self.step();
                loop {
                    match self.current {
                        Some(ch @ ('0'..='9' | '_')) => {
                            value.push(ch);
                            self.step();
                        }
                        _ => {
                            let value = value.replace('_', "").parse::<i64>().unwrap();
                            break self.token_with_value(TokenKind::Int, TokenValue::Int(value));
                        }
                    }
                }
            }
            Some('"') => {
                self.step();
                let mut value = String::new();
                loop {
                    match self.current {
                        Some('\\') => {
                            self.step();
                            match self.current {
                                Some('n') => value.push('\n'),
                                Some('r') => value.push('\r'),
                                Some('t') => value.push('\t'),
                                Some('0') => value.push('\0'),
                                Some(ch) => value.push(ch),
                                None => {
                                    self.error("malformed string");
                                    break self.token(TokenKind::Error);
                                }
                            }
                        }
                        Some('"') => {
                            self.step();
                            break self
                                .token_with_value(TokenKind::Str, TokenValue::Str(value));
                        }
                        Some(ch) => {
                            value.push(ch);
                        }
                        _ => {
                            self.error("malformed string");
                            break self.token(TokenKind::Error);
                        }
                    }
                    self.step()
                }
            }
            Some('/') => {
                self.step();
                match self.current {
                    Some('/') => {
                        self.step();
                        loop {
                            match self.current {
                                None | Some('\n') => break self.next_token(),
                                _ => {
                                    self.step();
                                }
                            }
                        }
                    }
                    Some('*') => {
                        self.step();
                        let mut lch = self.current;
                        loop {
                            match (lch, self.current) {
                                (Some('*'), Some('/')) => break self.next_token(),
                                (_, Some(ch)) => {
                                    lch = Some(ch);
                                    self.step();
                                }
                                (_, None) => {
                                    self.error("malformed /**/ comment");
                                    break self.token(TokenKind::Error);
                                }
                            }
                        }
                    }
                    _ => todo!(),
                }
            }
            Some('-') => {
                self.step();
                match self.current {
                    Some('>') => {
                        self.step();
                        self.token(TokenKind::MinusLt)
                    }
                    _ => todo!(),
                }
            }
            Some(ch @ ('0' | '(' | ')' | '{' | '}' | ':' | ',' | ';' | '=')) => {
                self.step();
                match ch {
                    '0' => self.token_with_value(TokenKind::Int, TokenValue::Int(0)),
                    '(' => self.token(TokenKind::LParen),
                    ')' => self.token(TokenKind::RParen),
                    '{' => self.token(TokenKind::LBrace),
                    '}' => self.token(TokenKind::RBrace),
                    ':' => self.token(TokenKind::Colon),
                    ',' => self.token(TokenKind::Comma),
                    ';' => self.token(TokenKind::Semicolon),
                    '=' => self.token(TokenKind::Equal),
                    _ => unreachable!(),
                }
            }
            Some(ch) => {
                self.error(format!("unknown char '{ch}'"));
                self.token(TokenKind::Error)
            }
        }
    }

    fn step(&mut self) {
        match self.current {
            Some('\n') => {
                self.line += 1;
                self.col = 1;
            }
            Some(_) => {
                self.col += 1;
            }
            _ => {}
        }
        self.current = self.chars.next();
        if self.current.is_some() {
            self.index += 1;
        }
    }

    fn token(&self, kind: TokenKind) -> Option<Token> {
        Some(Token {
            kind,
            value: TokenValue::None,
            index: self.index,
            line: self.line,
            col: self.col,
        })
    }

    fn token_with_value(&self, kind: TokenKind, value: TokenValue) -> Option<Token> {
        Some(Token {
            kind,
            value,
            index: self.index,
            line: self.line,
            col: self.col,
        })
    }

    fn error<S: Into<String>>(&mut self, msg: S) {
        let msg = msg.into();
        println!("lexer error: {msg}, line {}", self.line)
    }

    fn done(&self) -> bool {
        self.current.is_none()
    }
}

impl Iterator for Lexer<'_> {
    type Item = Token;

    fn next(&mut self) -> Option<Self::Item> {
        self.next_token()
    }
}

#[test]
fn test_lexer() {
    use TokenKind as TK;
    use TokenValue as TV;

    let lex = |text| {
        Lexer::new(text)
            .map(|token| (token.kind, token.value))
            .collect::<Vec<_>>()
    };

    assert_eq!(lex("abc"), vec![(TK::Id, TV::Id(hash("abc")))]);
    assert_eq!(lex("123"), vec![(TK::Int, TV::Int(123))]);
    assert_eq!(lex("\"\""), vec![(TK::Str, TV::Str("".to_string()))]);
    assert_eq!(
        lex("\"hello\""),
        vec![(TK::Str, TV::Str("hello".to_string()))]
    );
    assert_eq!(
        lex("\"new\\nline\""),
        vec![(TK::Str, TV::Str("new\nline".to_string()))]
    );
    assert_eq!(
        lex("\"backslash\\\\\""),
        vec![(TK::Str, TV::Str("backslash\\".to_string()))]
    );
    assert_eq!(lex("->"), vec![(TK::MinusLt, TV::None)]);
    assert_eq!(lex("let"), vec![(TK::Let, TV::None)]);
}
