use std::collections::HashMap;

use crate::{
    lexer::Lexer,
    parsed::Node,
    token::{Token, TokenKind, TokenValue},
};

pub struct Parser<'a> {
    lexer: Lexer<'a>,
    current: Option<Token>,
}

impl<'a> Parser<'a> {
    pub fn new(text: &'a str) -> Self {
        let mut lexer = Lexer::new(text);
        let current = lexer.next();
        Self { lexer, current }
    }

    pub fn parse(&mut self) -> Vec<Node> {
        self.parse_file()
    }

    pub fn symbols(self) -> HashMap<u64, String> {
        self.lexer.symbols()
    }

    fn parse_file(&mut self) -> Vec<Node> {
        let mut stmts = Vec::new();
        loop {
            match self.current {
                Some(_) => stmts.push(self.parse_stmt()),
                None => break stmts,
            }
        }
    }

    fn parse_stmt(&mut self) -> Node {
        match self.curr_kind() {
            Some(TokenKind::LBrace) => self.parse_block(),
            Some(TokenKind::If) => self.parse_if(),
            Some(TokenKind::Loop) => self.parse_loop(),
            Some(TokenKind::Fn) => self.parse_fn(),
            _ => {
                let stmt = match self.curr_kind() {
                    Some(TokenKind::Let) => self.parse_let(),
                    Some(TokenKind::Break) => {
                        self.step();
                        Node::Break
                    }
                    Some(TokenKind::Return) => {
                        self.step();
                        let value = match self.curr_kind() {
                            Some(TokenKind::Semicolon) => None,
                            _ => Some(Box::new(self.parse_expr())),
                        };
                        Node::Return { value }
                    }
                    _ => self.parse_assign(),
                };
                match self.curr_kind() {
                    Some(TokenKind::Semicolon) => {
                        self.step();
                        stmt
                    }
                    _ => {
                        self.error("expected ';'");
                        Node::Error
                    }
                }
            }
        }
    }

    fn parse_fn(&mut self) -> Node {
        self.step();
        if !self.curr_is(TokenKind::Id) {
            self.error("expected id");
            return Node::Error;
        }
        let subject = Box::new(self.parse_id());
        if !self.curr_is(TokenKind::LParen) {
            self.error("expected '('");
            return Node::Error;
        }
        let params = match self.parse_fn_params() {
            Ok(params) => params,
            Err(expr) => return expr,
        };
        if !self.curr_is(TokenKind::MinusLt) {
            self.error("expected '->'");
            return Node::Error;
        }
        self.step();
        let return_typ = Box::new(self.parse_typ());
        if !self.curr_is(TokenKind::LBrace) {
            self.error("expected '{'");
            return Node::Error;
        }
        let body = Box::new(self.parse_block());
        Node::Fn {
            subject,
            params,
            return_typ,
            body,
        }
    }

    fn parse_fn_params(&mut self) -> Result<Vec<Node>, Node> {
        self.step();
        let mut params = Vec::new();
        if !self.curr_is(TokenKind::RParen) {
            if !self.curr_is(TokenKind::Id) {
                self.error("expected id");
                return Err(Node::Error);
            }
            params.push(self.parse_param());
            while let Some(TokenKind::Comma) = self.curr_kind() {
                self.step();
                if self.curr_is(TokenKind::RParen) {
                    self.error("expected ')'");
                    break;
                }
                params.push(self.parse_param());
            }
        }
        if !self.curr_is(TokenKind::RParen) {
            self.error("expected ')'");
            return Err(Node::Error);
        }
        self.step();
        Ok(params)
    }

    fn parse_let(&mut self) -> Node {
        self.step();
        if !self.curr_is(TokenKind::Id) {
            self.error("expected id");
            return Node::Error;
        }
        let subject = self.parse_param();
        if !self.curr_is(TokenKind::Equal) {
            self.error("expected '='");
            return Node::Error;
        }
        self.step();
        let value = self.parse_expr();
        Node::Let {
            subject: Box::new(subject),
            value: Box::new(value),
        }
    }

    fn parse_param(&mut self) -> Node {
        let subject = Box::new(self.parse_id());
        let typ = if let Some(TokenKind::Colon) = self.curr_kind() {
            self.step();
            Some(Box::new(self.parse_typ()))
        } else {
            None
        };
        Node::Param { subject, typ }
    }

    fn parse_typ(&mut self) -> Node {
        match self.curr_kind() {
            Some(TokenKind::Id) => self.parse_id(),
            _ => {
                self.error("expected type");
                self.step();
                Node::Error
            }
        }
    }

    fn parse_assign(&mut self) -> Node {
        let subject = self.parse_expr();
        match self.curr_kind() {
            Some(TokenKind::Equal) => {
                self.step();
                let value = self.parse_expr();
                Node::Assign {
                    subject: Box::new(subject),
                    value: Box::new(value),
                }
            }
            _ => subject,
        }
    }

    fn parse_expr(&mut self) -> Node {
        self.parse_call()
    }

    fn parse_call(&mut self) -> Node {
        let mut subject = self.parse_value();
        loop {
            match self.curr_kind() {
                Some(TokenKind::LParen) => {
                    self.step();
                    let mut args = Vec::new();
                    match self.curr_kind() {
                        None | Some(TokenKind::RParen) => {}
                        Some(_) => {
                            args.push(self.parse_expr());
                            while let Some(TokenKind::Comma) = self.curr_kind() {
                                self.step();
                                if let Some(TokenKind::RParen) = self.curr_kind() {
                                    break;
                                }
                                args.push(self.parse_expr());
                            }
                        }
                    }
                    match self.curr_kind() {
                        Some(TokenKind::RParen) => {}
                        _ => {
                            self.error("expected ')'");
                            return Node::Error;
                        }
                    }
                    self.step();
                    subject = Node::Call {
                        subject: Box::new(subject),
                        args,
                    };
                }
                _ => break subject,
            }
        }
    }

    fn parse_value(&mut self) -> Node {
        match self.curr_kind() {
            Some(TokenKind::Id) => self.parse_id(),
            Some(TokenKind::Int) => self.parse_int(),
            Some(TokenKind::Str) => self.parse_string(),
            Some(TokenKind::LParen) => self.parse_group(),
            Some(TokenKind::LBrace) => self.parse_block(),
            Some(TokenKind::If) => self.parse_if(),
            Some(TokenKind::Loop) => self.parse_loop(),
            _ => {
                self.error("expected value");
                self.step();
                Node::Error
            }
        }
    }

    fn parse_id(&mut self) -> Node {
        let Some(Token {
                kind: TokenKind::Id,
                value: TokenValue::Id(value),
                ..
            }) = self.current
            else {
                unreachable!()
            };
        self.step();
        Node::Id(value)
    }

    fn parse_int(&mut self) -> Node {
        let Some(Token {
                kind: TokenKind::Int,
                value: TokenValue::Int(value),
                ..
            }) = self.current
            else {
                unreachable!()
            };
        self.step();
        Node::Int(value)
    }

    fn parse_string(&mut self) -> Node {
        let Some(Token {
                kind: TokenKind::Str,
                    value: TokenValue::Str(value),
                    ..
                }) = self.current.clone()
                else {
                    unreachable!()
                };
        self.step();
        Node::Str(value.clone())
    }

    fn parse_group(&mut self) -> Node {
        self.step();
        let expr = Box::new(self.parse_expr());
        if !self.curr_is(TokenKind::RParen) {
            self.error("expected ')'");
            return Node::Error;
        }
        self.step();
        Node::Group(expr)
    }

    fn parse_block(&mut self) -> Node {
        self.step();
        let mut stmts = Vec::new();
        loop {
            match self.curr_kind() {
                None => {
                    self.error("expected ')'");
                    break Node::Error;
                }
                Some(TokenKind::RBrace) => {
                    self.step();
                    break Node::Block(stmts);
                }
                _ => stmts.push(self.parse_stmt()),
            }
        }
    }

    fn parse_if(&mut self) -> Node {
        self.step();
        let cond = Box::new(self.parse_expr());
        if !self.curr_is(TokenKind::LBrace) {
            self.error("expected '}'");
            return Node::Error;
        }
        let truthy = Box::new(self.parse_block());
        let falsy = match self.curr_kind() {
            Some(TokenKind::Else) => {
                self.step();
                if !self.curr_is(TokenKind::LBrace) {
                    self.error("expected '}'");
                    return Node::Error;
                }
                Some(Box::new(self.parse_block()))
            }
            _ => None,
        };
        Node::If {
            cond,
            truthy,
            falsy,
        }
    }

    fn parse_loop(&mut self) -> Node {
        self.step();
        if !self.curr_is(TokenKind::LBrace) {
            self.error("expected '}'");
            return Node::Error;
        }
        let body = Box::new(self.parse_block());
        Node::Loop { body }
    }

    fn error<S: Into<String>>(&mut self, msg: S) {
        let msg = msg.into();
        println!(
            "parser error: {msg}, line {}",
            self.current
                .as_ref()
                .map(|t| t.line.to_string())
                .unwrap_or_else(|| "-".to_string())
        )
    }

    fn step(&mut self) {
        self.current = self.lexer.next();
    }

    fn curr_is(&self, kind: TokenKind) -> bool {
        self.curr_kind() == Some(kind)
    }

    fn curr_kind(&self) -> Option<TokenKind> {
        self.current.as_ref().map(|t| t.kind.clone())
    }
}

#[test]
fn test_parser() {
    use crate::util::hash;
    use Node::*;

    let parse = |text| Parser::new(text).parse();

    #[allow(non_snake_case)]
    fn B<T>(v: T) -> Box<T> {
        Box::new(v)
    }

    assert_eq!(Parser::new("abc;").parse(), vec![Id(hash("abc"))]);
    assert_eq!(Parser::new("123;").parse(), vec![Int(123)]);
    assert_eq!(
        Parser::new("\"hello\";").parse(),
        vec![Str("hello".to_string())]
    );
    assert_eq!(Parser::new("0;").parse(), vec![Int(0)]);
    assert_eq!(Parser::new("0;abc;").parse(), vec![Int(0), Id(hash("abc"))]);
    assert_eq!(
        parse("add(mul(12, 34), 56);"),
        vec![Call {
            subject: B(Id(hash("add"))),
            args: vec![
                Call {
                    subject: B(Id(hash("mul"))),
                    args: vec![Int(12), Int(34)]
                },
                Int(56)
            ]
        }]
    );
    assert_eq!(
        parse("a = 123;"),
        vec![Assign {
            subject: B(Id(hash("a"))),
            value: B(Int(123))
        }]
    );
    assert_eq!(parse("break;"), vec![Break]);
    assert_eq!(parse("return;"), vec![Return { value: None }]);
    assert_eq!(
        parse("return add(1, 2);"),
        vec![Return {
            value: Some(B(Call {
                subject: B(Id(hash("add"))),
                args: vec![Int(1), Int(2)]
            }))
        }]
    );
    assert_eq!(
        parse("a = 5;"),
        vec![Assign {
            subject: B(Id(hash("a"))),
            value: B(Int(5))
        }]
    );
    assert_eq!(
        parse("let a = 5;"),
        vec![Let {
            subject: B(Param {
                subject: B(Id(hash("a"))),
                typ: None
            }),
            value: B(Int(5))
        }]
    );
    assert_eq!(
        parse("fn test() -> i32 {}"),
        vec![Fn {
            subject: B(Id(hash("test"))),
            params: vec![],
            return_typ: B(Id(hash("i32"))),
            body: B(Block(vec![]))
        }]
    );
    assert_eq!(
        parse("if 0 {}"),
        vec![If {
            cond: B(Int(0)),
            truthy: B(Block(vec![])),
            falsy: None
        }]
    );
    assert_eq!(
        parse("if 0 {} else {}"),
        vec![If {
            cond: B(Int(0)),
            truthy: B(Block(vec![])),
            falsy: Some(B(Block(vec![]))),
        }]
    );
    assert_eq!(
        parse("loop {}"),
        vec![Loop {
            body: B(Block(vec![])),
        }]
    );
}
